"""
Classes to handle oauth portion of LTI
"""
# pylint: disable=C0103
import logging

import oauth.oauth as oauth

from xsiftx.config import settings, get_consumer

log = logging.getLogger('xsiftx')


class LTIOAuthDataStore(oauth.OAuthDataStore):
    """
    Largely taken from reference implementation
    for app engine at https://code.google.com/p/ims-dev/
    """

    def __init__(self):
        """
        Create OAuth store
        """
        oauth.OAuthDataStore.__init__(self)
        self.consumers = settings.get('consumers', None)

    def lookup_consumer(self, key):
        """
        Search through keys
        """
        if not self.consumers:
            log.critical(("No consumers defined in settings."
                          "Have you created a configuration file?"))
            return None

        consumer = get_consumer(key)
        if not consumer:
            log.info("Did not find consumer, using key: %s ", key)
            return None

        secret = consumer.get('secret', None)
        if not secret:
            return None
        return oauth.OAuthConsumer(key, secret)

    def lookup_token(self, oauth_consumer, token_type, token):
        """We don't do request_tokens"""
        # pylint: disable=W0613
        return oauth.OAuthToken(None, None)

    def lookup_nonce(self, oauth_consumer, oauth_token, nonce):
        """Trust all nonces"""
        return None  # pragma: no cover

    def fetch_request_token(self, oauth_consumer, oauth_callback):
        """We don't do request_tokens"""
        return None  # pragma: no cover

    def fetch_access_token(self, oauth_consumer, oauth_token, oauth_verifier):
        """We don't do request_tokens"""
        return None  # pragma: no cover

    def authorize_request_token(self, oauth_token, user):
        """We don't do request_tokens"""
        return None  # pragma: no cover



"""
Test utilities
"""

import contextlib
import shutil
import sys
import tempfile


@contextlib.contextmanager
def nostderr():
    """
    Replace standard error with nothing
    as a context manager so it is restored properly
    """
    # pylint: disable=r0903
    savestderr = sys.stderr

    class Devnull(object):
        """
        Fake dev null by skipping write operations.
        """
        def write(self, _):
            """
            Just pass on write
            """
            pass
    sys.stderr = Devnull()
    try:
        yield
    finally:
        sys.stderr = savestderr


def mkdtemp_clean(test_class):
    """
    Make a temp directory and add a cleanup action to it
    """
    temp_dir = tempfile.mkdtemp()
    test_class.addCleanup(shutil.rmtree, temp_dir)
    return temp_dir



"""
This handles getting the application configuration such
as the OAuth key list and what sifters are authorized
for what key.

This uses yaml as the configuration format and tries to load the
configuration from $(pwd)/xsiftx.yml then ~/.xsiftx.yml, and
then /etc/xsiftx.yml using the first one it finds. It can also
be specified with the environment variable XSIFTX_CONFIG set to a
path.

The format is:

edx_venv_path: "/edx/app/edxapp/venvs/edxapp"
edx_platform_path: "/edx/app/edxapp/edx-platform"
consumers:
  - key: <lti_app_identifier. e.g MITx-6.00x>
    secret: <oath_secret>
    allowed_sifters:
      - dump_grades
      - ...

if sifter list is empty or unspecified, all sifters are allowed. It
also requires a server secret in order to use secure client cookies to
store session information as flask_secret_key: <long crypto secret>
"""
import os
import yaml

CONFIG_PATHS = [
    os.path.join(os.getcwd(), 'xsiftx.yml'),
    os.path.join(os.path.expanduser('~'), '.xsiftx.yml'),
    '/etc/xsiftx.yml',
]

VENV = ('edx_venv_path', '/edx/app/edxapp/venvs/edxapp')
EDX_PLATFORM = ('edx_platform_path', '/edx/app/edxapp/edx-platform')


def get_consumer(key):
    """
    Returns the consumer object based on key
    """
    consumers = settings.get('consumers', None)
    consumer = next(
        (consumer for consumer in consumers
         if consumer.get("key", None) == key),
        None
    )
    return consumer


def get_config():
    """
    Find config file and load or return None
    """
    conf = None
    config_file = None

    var_conf = os.environ.get('XSIFTX_CONFIG', None)
    print CONFIG_PATHS
    if var_conf:
        config_file = var_conf
    else:
        for conf_path in CONFIG_PATHS:
            print(conf_path)
            if os.path.isfile(conf_path):
                config_file = conf_path
                break

    if config_file:
        with open(config_file) as conf_yaml:
            conf = yaml.load(conf_yaml)

    if not conf:
        raise Exception('No configuration found')

    # Add some defaults for common settings if they aren't there
    if not conf.get(VENV[0], None):
        conf[VENV[0]] = VENV[1]

    if not conf.get(EDX_PLATFORM[0], None):
        conf[EDX_PLATFORM[0]] = EDX_PLATFORM[1]

    return conf

settings = get_config()  # pylint: disable=C0103



"""
This is a simple LTI interface application/blueprint
for flask that allows user's to run a sifter from
inside their courseware (if authorized)
"""
# pylint: disable=C0103
import os
import shlex
import time

from celery import Celery
from flask import (
    Blueprint,
    render_template,
    session,
    request,
    jsonify,
    make_response,
)
from jinja2 import TemplateNotFound

from .decorators import lti_authentication, lti_staff_required
from .util import (
    InvalidAPIUsage,
    LTIException,
    LTIRoleException,
    get_allowed_sifters
)
from xsiftx.config import settings, get_consumer, VENV, EDX_PLATFORM
from xsiftx.util import (
    get_sifters,
    XsiftxException,
    SifterException,
    run_sifter
)


API_VERSION = 'v0.1'

JOB_CLEAR_STATUSES = ['SUCCESS', 'FAILURE', 'REVOKED', 'SIFTER_FAILURE', ]


# Define our app as a blueprint
xsiftx_lti = Blueprint(
    'xsiftx_lti',
    __name__,
    template_folder='templates/lti',
    static_folder='static/lti',
)


celery = Celery(
    'xsiftx_lti',
    broker=settings.get(
        'CELERY_BROKER_URL',
        'amqp://guest:guest@127.0.0.1:5672/'
    ),
    backend=settings.get(
        'CELERY_RESULT_BACKEND', 'amqp'
    )
)
celery.conf.update(settings)


@xsiftx_lti.route('/', methods=['GET', 'POST'])
@lti_authentication
@lti_staff_required
def index():
    """
    Show the xsift interface page
    """
    consumer = get_consumer(session['oauth_consumer_key'])
    sifters = get_allowed_sifters(consumer)

    return render_template(
        'index.html',
        sifters=sifters,
    )


@xsiftx_lti.route('/api/{0}/run'.format(API_VERSION), methods=['POST'])
@lti_authentication
def run():
    """
    Runs a given sifter for the course in the LTI component.
    """
    sifter_name = request.form.get('sifter', None)
    if not sifter_name:
        raise InvalidAPIUsage(
            'This api call requires the parameter "sifter".'
        )
    consumer = get_consumer(session['oauth_consumer_key'])
    sifters = get_allowed_sifters(consumer, True)
    sifter = sifters.get(sifter_name, None)
    if not sifter:
        raise InvalidAPIUsage(
            'You have specified an invalid sifter.',
            400,
            {'available_sifters': sifters}
        )

    course = session['context_id']
    extra_args = shlex.split(request.form.get('extra_args', ''))
    task = web_run_sifter.apply_async((
        sifter,
        course,
        extra_args
    ))
    managed_tasks = list(session.get('managed_tasks', []))
    task_dict = {
        'sifter': sifter_name,
        'task_id': task.task_id,
        'time': time.strftime('%Y-%m-%d %H:%M:%SZ', time.localtime()),
        'extra_args': extra_args,
        'course': course,
    }
    managed_tasks.append(task_dict)
    session['managed_tasks'] = managed_tasks
    return get_task_status()


@xsiftx_lti.route(
    '/api/{0}/update_task_status'.format(API_VERSION),
    methods=['PUT']
)
@lti_authentication
def get_task_status():
    """
    Grabs a status of all the tasks that are stored in the session
    """
    managed_tasks = session.get('managed_tasks', [])
    for task in managed_tasks:
        result = celery.AsyncResult(task['task_id'])  # pylint: disable=e1121
        task['status'] = result.status
        if result.state == 'SUCCESS':
            task['results'] = result.result
            if not task['results']['success']:
                task['status'] = 'SIFTER_FAILURE'
    session['managed_tasks'] = managed_tasks
    return jsonify({'tasks': managed_tasks})


@xsiftx_lti.route(
    '/api/{0}/clear_complete_tasks'.format(API_VERSION),
    methods=['DELETE'])
@lti_authentication
def clear_complete_tasks():
    """
    Nukes completed tasks from the session store
    """
    # Update list
    get_task_status()
    managed_tasks = session.get('managed_tasks', [])
    managed_tasks[:] = [
        task for task in managed_tasks
        if not task['status'] in JOB_CLEAR_STATUSES
    ]
    return jsonify({'tasks': managed_tasks})


@celery.task(name='xsiftx.run_sifter')
def web_run_sifter(sifter, course, extra_args):
    """
    Run the given sifter and handle errors from the internal call
    """
    error = u''
    success = True
    try:
        run_sifter(
            sifter,
            course,
            settings[VENV[0]],
            settings[EDX_PLATFORM[0]],
            extra_args
        )
    except XsiftxException as err:
        error = unicode(err)
        success = False
    except SifterException as err:
        error = unicode(err)
        success = False

    return {
        'success': success,
        'sifter': os.path.basename(sifter),
        'error': error,
    }


@xsiftx_lti.errorhandler(InvalidAPIUsage)
def handle_invalid_api_usage(error):
    """Return jsonified error message"""
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    return response


@xsiftx_lti.errorhandler(LTIException)
@xsiftx_lti.errorhandler(LTIRoleException)
def handle_lti_error(error):
    """Return LTI error"""
    # Return json if this is from an API call or html
    # if not.
    if '/api/{0}'.format(API_VERSION) in request.url:
        response = jsonify({'message': unicode(error)})
    else:
        response = make_response(render_template(
            'lti_error.html',
            message=unicode(error)
        ))
    response.status_code = 401
    return response



"""
Flask application entry for xsiftx LTI
provider application.
"""
from flask import Flask

from xsiftx.config import settings
from xsiftx.lti import xsiftx_lti


app = Flask('xsiftx')  # pylint: disable=C0103
app.config.update(**settings)  # pylint: disable=W0142


# Setup session from config
app.secret_key = settings.get('flask_secret_key', None)

# Register blueprints
app.register_blueprint(xsiftx_lti)

if __name__ == "__main__":
    app.debug = True
    app.run(host='0.0.0.0')



"""
Tests for xsiftx.util functions
"""
import os
import stat
import unittest

from mock import patch

from .util import mkdtemp_clean
from xsiftx.util import (
    get_sifters,
    get_course_list,
    get_settings,
    run_sifter,
    XsiftxException,
    SifterException
)


class TestUtils(unittest.TestCase):
    """
    Test series for util functions in xsiftx.util
    """
    # pylint: disable=r0904

    KNOWN_SIFTERS = [
        'copy_file',
        'xqanalyze',
        'content_statistics',
        'test_sifters',
        'dump_grades',
    ]

    EDX_ROOT = '/edx/app/edxapp/edx-platform'
    EDX_VENV = '/edx/app/edxapp/venvs/edxapp'

    BAD_SIFTER = 'testenv_sifter'

    def _make_bad_sifter(self):
        """
        Create a sifter that raises an exception
        """
        temp_dir = mkdtemp_clean(self)
        sifter_path = os.path.join(temp_dir, self.BAD_SIFTER)
        with open(sifter_path, 'w+') as temp_sifter:
            temp_sifter.write('#!/bin/bash\nfalse')
        perms = os.stat(sifter_path)
        os.chmod(sifter_path, perms.st_mode | stat.S_IEXEC)
        os.environ['SIFTER_DIR'] = temp_dir

    def test_sifter_list_locations(self):
        """
        Make sure sifter search paths work
        """
        # check default sifters
        self.assertTrue(set(self.KNOWN_SIFTERS).issubset(get_sifters()))

        # Add environment variable of extra sifter folder, ensure it is added
        # to list
        self._make_bad_sifter()
        self.assertTrue(self.BAD_SIFTER in get_sifters())

    @unittest.skipUnless(os.environ.get('XSIFTX_TEST_EDX', None),
                         'Requires an edx environment and XSIFTX_TEST_EDX '
                         'environment variable set.')
    def test_course_list(self):
        """
        Make sure we can get a course list.
        Requires that we have an edx-platform environment at default
        coords
        """
        with self.assertRaisesRegexp(XsiftxException,
                                     'No such file or dir.*'):
            get_course_list('nope', 'nope')

        self.assertTrue(len(get_course_list(self.EDX_VENV, self.EDX_ROOT)) > 0)

    @unittest.skipUnless(os.environ.get('XSIFTX_TEST_EDX', None),
                         'Requires an edx environment and XSIFTX_TEST_EDX '
                         'environment variable set.')
    def test_get_settings(self):
        """
        Test that we can get settings and fail nicely without them
        """
        settings_keys = [
            'use_s3',
            'aws_key',
            'root_path',
            'bucket',
            'aws_key_id',
        ]
        with self.assertRaisesRegexp(XsiftxException,
                                     'Cannot find lms.*'):
            get_settings('nope')
        settings = get_settings(self.EDX_ROOT)
        self.assertEqual(settings_keys, settings.keys())
        self.assertIsNotNone(settings.get('root_path'))
        self.assertIsNotNone(settings.get('bucket'))

    @unittest.skipUnless(os.environ.get('XSIFTX_TEST_EDX', None),
                         'Requires an edx environment and XSIFTX_TEST_EDX '
                         'environment variable set.')
    def test_run_sifter(self):
        """
        Run sifters to make sure the runner is working right
        """
        # Try with bad root
        with self.assertRaisesRegexp(XsiftxException,
                                     'Cannot find lms*'):
            run_sifter('nope', 'nope', 'nope', 'nope', 'nope')

        # Mock out settings so we know where output is going
        bucket = 'reports'
        course = 'stuff'
        temp_dir = mkdtemp_clean(self)

        with patch('xsiftx.util.get_settings') as mock_settings:
            mock_settings.return_value = {
                'use_s3': False,
                'aws_key': '',
                'root_path': temp_dir,
                'bucket': bucket,
                'aws_key_id': ''
            }
            # Test that a bad sifter raises
            with self.assertRaisesRegexp(SifterException,
                                         'Sifter .+ called with'):
                self._make_bad_sifter()
                run_sifter(get_sifters()[self.BAD_SIFTER],
                           course, self.EDX_VENV, self.EDX_ROOT, [])

            # Test that a good sifter creates the expected file
            run_sifter(get_sifters()['test_sifters'],
                       course, self.EDX_VENV, self.EDX_ROOT, [])
            self.assertTrue(mock_settings.called)
            self.assertTrue(os.path.exists(
                os.path.join(temp_dir, course, 'test_sifter.txt')
            ))



"""
Decorators to handle LTI session management,
authentication, etc.
"""
# pylint: disable=C0103
from functools import wraps
import logging

from flask import request, session
import oauth.oauth as oauth

from .oauthstore import LTIOAuthDataStore
from .util import LTIException, LTIRoleException

log = logging.getLogger('xsiftx')

LTI_PROPERTY_LIST = [
    'oauth_consumer_key',
    'launch_presentation_return_url',
    'user_id',
    'oauth_nonce',
    'context_label',
    'context_id',
    'resource_link_title',
    'resource_link_id',
    'lis_person_contact_email_primary',
    'lis_person_contact_emailprimary',
    'lis_person_name_full',
    'lis_person_name_family',
    'lis_person_name_given',
    'lis_result_sourcedid',
    'launch_type',
    'lti_message',
    'lti_version',
    'roles',
]

LTI_STAFF_ROLES = ['Instructor', 'Administrator', ]

LTI_SESSION_KEY = 'lti_authenticated'


def lti_authentication(func):
    """
    This is a middleware to handle LTI session
    and authorization for a given view.
    """
    @wraps(func)
    def decorator(*args, **kwargs):
        """
        Actual wrapper to handle LTI/OAuth
        """
        # pylint: disable=W0212
        # If we are already authenticated, just
        # return instead
        if session.get(LTI_SESSION_KEY, False):
            return func(*args, **kwargs)

        # Get lti GET or POST params as dict
        if request.method == 'POST':
            params = request.form.to_dict()
        else:
            params = request.args.to_dict()

        log.debug(params)

        # Try and authentication if session is being initiated
        oauth_server = oauth.OAuthServer(LTIOAuthDataStore())
        oauth_server.add_signature_method(
            oauth.OAuthSignatureMethod_PLAINTEXT())
        oauth_server.add_signature_method(
            oauth.OAuthSignatureMethod_HMAC_SHA1())

        oauth_request = oauth.OAuthRequest.from_request(
            "POST",
            request.url,
            headers=dict(request.headers),
            parameters=params
        )

        if not oauth_request:
            log.info('Received non oauth request on oauth protected page')
            raise LTIException('This page requires a valid oauth session '
                               'or request')
        try:
            consumer = oauth_server._get_consumer(oauth_request)
            oauth_server._check_signature(oauth_request, consumer, None)
        except oauth.OAuthError as err:
            # Rethrow our own for nice error handling (don't print
            # error message as it will contain the key
            log.info(err.message)
            raise LTIException("OAuth error: Please check your key and secret")

        # All good to go, store all of the LTI params into a
        # session dict for use in views
        for prop in LTI_PROPERTY_LIST:
            if params.get(prop, None):
                session[prop] = params[prop]

        # Set logged in session key
        session[LTI_SESSION_KEY] = True

        return func(*args, **kwargs)

    return decorator


def lti_staff_required(func):
    """
    Decorator to make sure that person is a
    member of one of the course staff roles
    before allowing them to the view. Requires that
    lti_authentication has occurred
    """
    @wraps(func)
    def decorator(*args, **kwargs):
        """
        Check session['role'] against known list of course staff
        roles and raise if it isn't in that set.
        """
        log.debug(session)
        role = session.get('roles', None)
        if not role:
            raise LTIRoleException(
                'User does not have a role. One is required'
            )
        if role not in LTI_STAFF_ROLES:
            raise LTIRoleException(
                'You are not in a staff level role. Access is restricted '
                'to course staff.'
            )
        return func(*args, **kwargs)

    return decorator



"""
Utility support functions
"""
from xsiftx.util import get_sifters


class LTIException(Exception):
    """
    Custom LTI exception for proper handling
    of LTI specific errors
    """
    pass


class LTIRoleException(Exception):
    """
    Exception class for when LTI user doesn't have the
    right role.
    """
    pass


class InvalidAPIUsage(Exception):
    """
    API Error handler to return helpful json when problems occur.
    Stolen right from the flask docs
    """
    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        """
        Setup class with optional arguments for returning later
        """
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        """
        Aggregate properties into dictionary for use
        in returning jsonified errors.
        """
        exception_dict = dict(self.payload or ())
        exception_dict['message'] = self.message
        return exception_dict


def get_allowed_sifters(consumer, as_dict=False):
    """
    Returns a list of sifter names allowed by the client
    """
    all_sifters = get_sifters()
    sifters = {}
    allowed_sifters = consumer.get('allowed_sifters', None)
    if allowed_sifters:
        for sifter in all_sifters.keys():
            if sifter in allowed_sifters:
                sifters[sifter] = all_sifters[sifter]
    else:
        sifters = all_sifters
    if not as_dict:
        return sifters.keys()
    else:
        return sifters



"""
Unit tests for xisftx command line interface
"""
import os
import sys
import unittest

from mock import patch

from .util import nostderr, mkdtemp_clean
from xsiftx.command_line import execute
from xsiftx.util import XsiftxException, get_course_list


class TestCommandLine(unittest.TestCase):
    """
    Test options of command line
    """
    # pylint: disable=R0904

    EDX_ROOT = '/edx/app/edxapp/edx-platform'
    EDX_VENV = '/edx/app/edxapp/venvs/edxapp'

    def test_args(self):
        """
        Test all the argument variations available
        """
        with nostderr():
            with self.assertRaises(SystemExit) as exception_context:
                execute()
            exit_exception = exception_context.exception
            self.assertEqual(exit_exception.code, 2)

            with self.assertRaises(SystemExit) as exception_context:
                sys.argv = ['xsiftx', 'sifter', ]
                execute()
            exit_exception = exception_context.exception
            self.assertEqual(exit_exception.code, 2)

            with self.assertRaises(SystemExit) as exception_context:
                sys.argv = ['xsiftx', '-v', 'sfutt', 'sifter', ]
                execute()
            exit_exception = exception_context.exception
            self.assertEqual(exit_exception.code, 2)

            with self.assertRaises(SystemExit) as exception_context:
                sys.argv = ['xsiftx', '-e', 'sfutt', 'sifter', ]
                execute()
            exit_exception = exception_context.exception
            self.assertEqual(exit_exception.code, 2)

    def test_bad_sifter(self):
        """
        Test exit code on non-existent sifter
        """
        with nostderr():
            with self.assertRaises(SystemExit) as exception_context:
                sys.argv = [
                    'xsiftx',
                    '-v', 'blah',
                    '-e', 'stuff',
                    'dontexist',
                ]
                execute()
            exit_exception = exception_context.exception
            self.assertEqual(exit_exception.code, -1)

    def test_bad_course_bad_env(self):
        """
        Test bad course
        """
        with self.assertRaisesRegexp(XsiftxException,
                                     'No such file or dir.*'):
            sys.argv = ['xsiftx',
                        '-v', 'blah',
                        '-e', 'stuff',
                        '-c', 'course',
                        'test_sifters', ]
            execute()

    @unittest.skipUnless(os.environ.get('XSIFTX_TEST_EDX', None),
                         'Requires an edx environment and XSIFTX_TEST_EDX '
                         'environment variable set.')
    def test_bad_course_good_env(self):
        """
        With proper parameters, test an invalid course
        """
        with nostderr():
            with self.assertRaises(SystemExit) as exception_context:
                sys.argv = ['xsiftx',
                            '-v', self.EDX_VENV,
                            '-e', self.EDX_ROOT,
                            '-c', 'not_a_course',
                            'test_sifters', ]
                execute()
            exit_exception = exception_context.exception
            self.assertEqual(exit_exception.code, -2)

    @unittest.skipUnless(os.environ.get('XSIFTX_TEST_EDX', None),
                         'Requires an edx environment and XSIFTX_TEST_EDX '
                         'environment variable set.')
    def test_specified_course(self):
        """
        With proper parameters, test specific course
        """
        temp_dir = mkdtemp_clean(self)
        with patch('xsiftx.util.get_settings') as mock_settings:
            mock_settings.return_value = {
                'use_s3': False,
                'aws_key': '',
                'root_path': temp_dir,
                'bucket': '',
                'aws_key_id': ''
            }
            sys.argv = ['xsiftx',
                        '-v', self.EDX_VENV,
                        '-e', self.EDX_ROOT,
                        '-c', get_course_list(self.EDX_VENV, self.EDX_ROOT)[0],
                        'test_sifters', ]
            execute()
            self.assertTrue(mock_settings.called)

    @unittest.skipUnless(os.environ.get('XSIFTX_TEST_EDX', None),
                         'Requires an edx environment and XSIFTX_TEST_EDX '
                         'environment variable set.')
    def test_valid_sifter_run(self):
        """
        Mock the location of the settings, but run valid sifter
        """
        temp_dir = mkdtemp_clean(self)
        with patch('xsiftx.util.get_settings') as mock_settings:
            mock_settings.return_value = {
                'use_s3': False,
                'aws_key': '',
                'root_path': temp_dir,
                'bucket': '',
                'aws_key_id': ''
            }
            sys.argv = ['xsiftx',
                        '-v', self.EDX_VENV,
                        '-e', self.EDX_ROOT,
                        'test_sifters', ]
            execute()
            self.assertTrue(mock_settings.called)



"""
Test LTI web components of the application
"""
import json
import os
import time
import unittest

import xsiftx.config
from xsiftx.config import get_config, get_consumer
from xsiftx.util import get_sifters
from xsiftx.lti.decorators import LTI_STAFF_ROLES
import xsiftx.web


class TestLTIWebApp(unittest.TestCase):
    """
    Test class for validating LTI interface
    to xsiftx.
    """
    # pylint: disable=e1103,r0904

    def setUp(self):
        """
        grab application test client
        """
        # pylint: disable=C0103
        self.client = xsiftx.web.app.test_client()
        self.settings = get_config()

    def _oauth_request(self, params={}, consumer_id=0):
        """
        Returns the data needed for an oauth request
        and adds those to the passed dictionary to
        keep out of the way.
        """
        # pylint: disable=w0102
        consumer = self.settings['consumers'][consumer_id]
        params.update({
            'oauth_consumer_key': consumer['key'],
            'oauth_signature_method': 'PLAINTEXT',
            'oauth_version': '1.0',
            'oauth_signature': '{0}&'.format(consumer['secret']),
            'oauth_timestamp': str(int(time.time())),
            'oauth_nonce': '123456789',
            'context_id': 'MITx/A.we/some'
        })
        return params

    def _run_sifter(self, sifter='test_sifters'):
        """
        Run a valid sifter and return json from response
        """
        test_url = 'api/v0.1/run'
        response = self.client.post(
            test_url,
            data=self._oauth_request({'sifter': sifter}, 1)
        )
        self.assertEqual(response.status_code, 200)
        return json.loads(response.data)

    def test_conf_settings(self):
        """
        Test and validate configuration and test rigging
        """
        self.assertEqual(
            self.settings['consumers'],
            [
                {'key': 'test_course1',
                 'secret': 'test_secret1',
                 'allowed_sifters': ['test_sifters']},
                {'key': 'test_course2',
                 'secret': 'test_secret2', },
            ]
        )

        # Test consumer search
        consumer = get_consumer('test_course1')
        self.assertEqual(consumer['secret'], 'test_secret1')

        # Test config exception
        conf_save = os.environ['XSIFTX_CONFIG']
        config_path_save = xsiftx.config.CONFIG_PATHS
        del os.environ['XSIFTX_CONFIG']

        xsiftx.config.CONFIG_PATHS = ['/dev/null', ]
        with self.assertRaisesRegexp(Exception, 'No configuration found'):
            settings = get_config()

        # Test config by path
        xsiftx.config.CONFIG_PATHS = [conf_save, ]
        settings = get_config()
        self.assertEqual(settings['awesome_canary_key'], 'test')

        os.environ['XSIFTX_CONFIG'] = conf_save
        xsiftx.config.CONFIG_PATHS = config_path_save

    def test_lti_authentication(self):
        """
        Test out that LTI authentication is working properly
        and that the decorator is raising the proper exceptions
        """
        # Setup empty request
        response = self.client.get('/')
        self.assertEqual(response.status_code, 401)
        self.assertTrue(
            'This page requires a valid oauth session or request' in
            response.data
        )

        # Test bad key
        params = self._oauth_request()
        params['oauth_consumer_key'] = 'silly'
        response = self.client.post('/', data=params)
        self.assertEqual(response.status_code, 401)
        self.assertTrue(
            'OAuth error: Please check your key' in
            response.data
        )

        # Test bad sig
        params = self._oauth_request()
        params['oauth_signature'] = 'silly&'
        response = self.client.post('/', data=params)
        self.assertEqual(response.status_code, 401)
        print(response.data)
        self.assertTrue(
            'OAuth error: Please check your key' in
            response.data
        )

        # Valid OAuth, but missing authorization
        response = self.client.post('/', data=self._oauth_request())
        self.assertEqual(response.status_code, 401)
        self.assertTrue(
            'User does not have a role' in
            response.data
        )

    def test_lti_staff_decorator(self):
        """
        Make sure that our authorization of role is happening
        """
        params = self._oauth_request({'roles': 'Student'})
        response = self.client.post('/', data=params)
        self.assertEqual(response.status_code, 401)
        self.assertTrue(
            'You are not in a staff level role. Access is restricted ' in
            response.data
        )

    def test_index(self):
        """
        Several tests of the index page
        """
        params = self._oauth_request({'roles': LTI_STAFF_ROLES[0]})
        response = self.client.post('/', data=params)
        self.assertEqual(response.status_code, 200)
        self.assertTrue(
            'Available Sifters' in
            response.data
        )

    def test_sifter_acl(self):
        """
        Test with consumer that has access to all sifters
        and ensure they all get loaded
        """
        # Test limited ACL
        with xsiftx.web.app.test_client() as client:
            consumer_id = 0
            params = self._oauth_request(
                {'roles': LTI_STAFF_ROLES[0]},
                consumer_id
            )
            response = client.post('/', data=params)
            self.assertEqual(response.status_code, 200)
            self.assertTrue(
                'Available Sifters' in
                response.data
            )
            # Make sure we don't have extra sifters
            self.assertTrue(
                all(sifter in response.data
                    for sifter in
                    self.settings['consumers'][consumer_id]['allowed_sifters'])
            )

        # Test no ACL
        with xsiftx.web.app.test_client() as client:
            consumer_id = 1
            params = self._oauth_request(
                {'roles': LTI_STAFF_ROLES[0]},
                consumer_id
            )
            response = client.post('/', data=params)
            self.assertEqual(response.status_code, 200)
            self.assertTrue(
                'Available Sifters' in
                response.data
            )
            # Make sure we don't have extra sifters
            self.assertTrue(
                all(sifter in response.data
                    for sifter in get_sifters())
            )

    def test_api_error(self):
        """
        Excercise the API error handler to make sure it returns
        json.
        """

        test_url = '/api/v0.1/run'

        # Test that LTI error on API url returns JSON instead of html
        response = self.client.post(test_url)
        self.assertEqual(response.status_code, 401)
        json.loads(response.data)

        # Test no sifter
        response = self.client.post(
            test_url,
            data=self._oauth_request()
        )
        self.assertEqual(response.status_code, 400)
        # Should raise if we don't get json
        reply_json = json.loads(response.data)
        self.assertTrue(
            u'This api call requires the parameter "sifter".' ==
            reply_json['message']
        )

        # Test invalid sifter
        response = self.client.post(
            test_url,
            data=self._oauth_request({'sifter': 'not_a_real_thing'})
        )
        self.assertEqual(response.status_code, 400)
        # Should raise if we don't get json
        reply_json = json.loads(response.data)
        self.assertTrue(
            u'You have specified an invalid sifter.' ==
            reply_json['message']
        )
        self.assertIsNotNone(reply_json.get('available_sifters', None))

    def test_run_sifter(self):
        """
        Test out the run command
        """
        reply_json = self._run_sifter()
        # Even though the job should fail, we should get
        # back the job id and such
        self.assertTrue(reply_json['tasks'][0]['status'] == 'PENDING')

    def test_task_status_and_delete(self):
        """
        Test task status API point and validate response json
        """
        update_url = '/api/v0.1/update_task_status'
        # Create several runs
        num_runs = 10
        for _ in range(num_runs):
            reply_json = self._run_sifter('xqanalyze')

        # Check we have the right number
        response = self.client.put(
            update_url,
            data=self._oauth_request()
        )
        reply_json = json.loads(response.data)
        self.assertTrue(len(reply_json['tasks']), num_runs)

        # Now delete them
        delete_url = '/api/v0.1/clear_complete_tasks'
        response = self.client.delete(
            delete_url,
            data=self._oauth_request()
        )
        reply_json = json.loads(response.data)
        self.assertTrue(len(reply_json['tasks']), 0)

        # And double check they are gone
        # Check we have the right number
        response = self.client.put(
            update_url,
            data=self._oauth_request()
        )
        reply_json = json.loads(response.data)
        self.assertTrue(len(reply_json['tasks']), 0)



"""
Utility functions for xsiftx.
"""
import json
import os
import stat
import subprocess
import tempfile

import xsiftx.sifters
import xsiftx.store

ENV_JSON_FILENAME = 'lms.env.json'
AUTH_JSON_FILENAME = 'lms.auth.json'


class XsiftxException(Exception):
    """
    Customized exception class for handling
    xsiftx errors like missing edx_platform
    configuration
    """
    pass


class SifterException(Exception):
    """
    Customized exception raised for sifter errors
    """
    pass


def get_sifters():
    """
    Get list of currently installed sifters
    """

    sifter_dict = {}
    # List of paths to look for sifters, ordered
    # in reverse precedence (most important last)
    # to replace sifter dictionary
    sifter_paths = [
        os.path.dirname(xsiftx.sifters.__file__),  # Installed sifters
        os.path.join(os.path.expanduser('~'), 'sifters'),  # HOME_DIR sifters
        os.path.join(os.getcwd(), 'sifters'),  # cwd sifters
        os.environ.get('SIFTER_DIR', ''),  # Environment set sifters
    ]

    executable = stat.S_IEXEC | stat.S_IXGRP | stat.S_IXOTH
    for sifter_path in sifter_paths:
        if os.path.isdir(sifter_path):
            for filename in os.listdir(sifter_path):
                fullpath = os.path.join(sifter_path, filename)
                if os.path.isfile(fullpath):
                    fstat = os.stat(fullpath)
                    mode = fstat.st_mode
                    if mode & executable:
                        sifter_dict[filename] = fullpath

    return sifter_dict


def get_course_list(venv, edx_root):
    """
    Get a list of courses by using the management commands in edx.
    """

    # Grab course list
    course_raw = ''
    try:
        course_raw = subprocess.check_output(
            ['{0}/bin/python'.format(venv),
             'manage.py', 'lms', '--settings=aws',
             'dump_course_ids', ],
            cwd=edx_root, stderr=subprocess.PIPE
        )
    except OSError as ex:
        raise XsiftxException(
            'No such file or directory: {0!r}\n'.format(str(ex))
        )

    except subprocess.CalledProcessError as ex:
        raise XsiftxException(
            'Course listing failed, output was: {0!r}\n'.format(ex.output)
        )
    return course_raw.split('\n')[:-1]


def get_settings(edx_root):
    """
    This will pull out the json settings for the
    platform in order to get the bucket, path, key_id, and key
    for uploading to the right place and return them as a dict.
    """
    auth_path = os.path.abspath('{0}/../{1}'.format(
        edx_root, AUTH_JSON_FILENAME))
    if not os.path.isfile(auth_path):
        raise XsiftxException(
            'Cannot find lms authentication file: {0}. '
            'Is the path correct?'.format(auth_path)
        )
    with open(auth_path) as auth_file:
        auth_tokens = json.load(auth_file)

    env_path = os.path.abspath('{0}/../{1}'.format(
        edx_root, ENV_JSON_FILENAME))
    if not os.path.isfile(env_path):
        raise XsiftxException(
            'Cannot find lms environment file: {0}. '
            'Is the path correct?'.format(env_path)
        )
    with open(env_path) as env_file:
        env_tokens = json.load(env_file)

    use_s3 = True
    grades_download = env_tokens['GRADES_DOWNLOAD']
    if grades_download['STORAGE_TYPE'] != 'S3':
        use_s3 = False

    bucket = grades_download['BUCKET']
    root_path = grades_download['ROOT_PATH']

    aws_key_id = auth_tokens['AWS_ACCESS_KEY_ID']
    if aws_key_id == '' and use_s3:
        raise XsiftxException('No AWS_ACCESS_KEY_ID\n')

    aws_key = auth_tokens["AWS_SECRET_ACCESS_KEY"]
    if aws_key == '' and use_s3:
        raise XsiftxException('No AWS_SECRET_ACCESS_KEY\n')

    return dict(aws_key=aws_key, aws_key_id=aws_key_id,
                bucket=bucket, root_path=root_path, use_s3=use_s3)


def run_sifter(sifter, course, venv, edx_platform, extra_args):
    """
    This handles running the actual sifter given a course
    and sifter
    """
    settings = get_settings(edx_platform)
    data_store = None

    if settings['use_s3']:
        data_store = xsiftx.store.S3Store(settings)
    else:
        data_store = xsiftx.store.FSStore(settings)

    with tempfile.NamedTemporaryFile() as tmpfile:
        with tempfile.NamedTemporaryFile() as stderr_tmp:
            cmd = [
                sifter,
                venv,
                edx_platform,
                course,
            ]
            cmd.extend(extra_args)
            sift = subprocess.Popen(cmd, stdout=tmpfile, stderr=stderr_tmp,
                                    universal_newlines=True)
            ret_code = sift.wait()
            if ret_code != 0:
                stderr_tmp.flush()
                stderr_tmp.seek(0)
                error_output = stderr_tmp.read()
                raise SifterException(
                    'Sifter {0} called with {1} failed '
                    'with non zero exit code printing output '
                    'and aborting\nError Output:\n{2}'.format(
                        sifter, ' '.join(cmd), error_output
                    )
                )

            tmpfile.flush()
            tmpfile.seek(0)
            if os.fstat(tmpfile.fileno()).st_size > 0:
                filename = tmpfile.readline()[:-1]
                data_store.store(course, filename, tmpfile)
